import textwrap
from random import random
from edsl.config import CONFIG

# if "EDSL_DEFAULT_MODEL" not in CONFIG:
#     default_model = "test"
# else:
#     default_model = CONFIG.get("EDSL_DEFAULT_MODEL")

from collections import UserList


class PrettyList(UserList):
    def __init__(self, data=None, columns=None):
        super().__init__(data)
        self.columns = columns

    def _repr_html_(self):
        if isinstance(self[0], list) or isinstance(self[0], tuple):
            num_cols = len(self[0])
        else:
            num_cols = 1

        if self.columns:
            columns = self.columns
        else:
            columns = list(range(num_cols))

        if num_cols > 1:
            return (
                "<pre><table>"
                + "".join(["<th>" + str(column) + "</th>" for column in columns])
                + "".join(
                    [
                        "<tr>"
                        + "".join(["<td>" + str(x) + "</td>" for x in row])
                        + "</tr>"
                        for row in self
                    ]
                )
                + "</table></pre>"
            )
        else:
            return (
                "<pre><table>"
                + "".join(["<th>" + str(index) + "</th>" for index in columns])
                + "".join(
                    ["<tr>" + "<td>" + str(row) + "</td>" + "</tr>" for row in self]
                )
                + "</table></pre>"
            )


def get_model_class(model_name, registry=None):
    from edsl.inference_services.registry import default

    registry = registry or default
    factory = registry.create_model_factory(model_name)
    return factory


class Meta(type):
    def __repr__(cls):
        return textwrap.dedent(
            f"""\
        Available models: {cls.available()}
        
        To create an instance, you can do: 
        >>> m = Model('gpt-4-1106-preview', temperature=0.5, ...)
        
        To get the default model, you can leave out the model name. 
        To see the available models, you can do:
        >>> Model.available()
        """
        )


class Model(metaclass=Meta):
    default_model = CONFIG.get("EDSL_DEFAULT_MODEL")

    def __new__(
        cls, model_name=None, registry=None, service_name=None, *args, **kwargs
    ):
        # Map index to the respective subclass
        if model_name is None:
            model_name = (
                cls.default_model
            )  # when model_name is None, use the default model, set in the config file
        from edsl.inference_services.registry import default

        registry = registry or default

        if isinstance(model_name, int):  # can refer to a model by index
            model_name = cls.available(name_only=True)[model_name]

        factory = registry.create_model_factory(model_name, service_name=service_name)
        return factory(*args, **kwargs)

    @classmethod
    def add_model(cls, service_name, model_name):
        from edsl.inference_services.registry import default

        registry = default
        registry.add_model(service_name, model_name)

    @classmethod
    def services(cls, registry=None):
        from edsl.inference_services.registry import default

        registry = registry or default
        return [r._inference_service_ for r in registry.services]

    @classmethod
    def available(cls, search_term=None, name_only=False, registry=None, service=None):
        from edsl.inference_services.registry import default

        registry = registry or default
        full_list = registry.available()

        if service is not None:
            if service not in cls.services(registry=registry):
                raise ValueError(f"Service {service} not found in available services.")

            full_list = [m for m in full_list if m[1] == service]

        if search_term is None:
            if name_only:
                return PrettyList(
                    [m[0] for m in full_list],
                    columns=["Model Name", "Service Name", "Code"],
                )
            else:
                return PrettyList(
                    full_list, columns=["Model Name", "Service Name", "Code"]
                )
        else:
            filtered_results = [
                m for m in full_list if search_term in m[0] or search_term in m[1]
            ]
            if name_only:
                return PrettyList(
                    [m[0] for m in filtered_results],
                    columns=["Model Name", "Service Name", "Code"],
                )
            else:
                return PrettyList(
                    filtered_results, columns=["Model Name", "Service Name", "Code"]
                )

    @classmethod
    def check_models(cls, verbose=False):
        print("Checking all available models...\n")
        for model in cls.available(name_only=True):
            print(f"Now checking: {model}")
            try:
                m = cls(model)
            except Exception as e:
                print(f"Error creating instance of {model}: {e}")
                continue
            try:
                results = m.hello(verbose)
                if verbose:
                    print(f"Results from model call: {results}")
            except Exception as e:
                print(f"Error calling 'hello' on {model}: {e}")
                continue
            print("OK!")
            print("\n")

    @classmethod
    def example(cls, randomize: bool = False) -> "Model":
        """
        Returns an example Model instance.

        :param randomize: If True, the temperature is set to a random decimal between 0 and 1.
        """
        temperature = 0.5 if not randomize else round(random(), 2)
        model_name = cls.default_model
        return cls(model_name, temperature=temperature)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    available = Model.available()
    m = Model("gpt-4-1106-preview")
    results = m.execute_model_call("Hello world")
    print(results)



from __future__ import annotations
import time
import asyncio
import threading
import warnings
from typing import Coroutine, List, AsyncGenerator, Optional, Union, Generator, Type
from uuid import UUID
from collections import UserList

from edsl.results.Results import Results
from edsl.jobs.interviews.Interview import Interview
from edsl.jobs.runners.JobsRunnerStatus import JobsRunnerStatus, JobsRunnerStatusBase

from edsl.jobs.tasks.TaskHistory import TaskHistory
from edsl.jobs.buckets.BucketCollection import BucketCollection
from edsl.utilities.decorators import jupyter_nb_handler
from edsl.data.Cache import Cache
from edsl.results.Result import Result
from edsl.results.Results import Results
from edsl.language_models.LanguageModel import LanguageModel
from edsl.data.Cache import Cache


class StatusTracker(UserList):
    def __init__(self, total_tasks: int):
        self.total_tasks = total_tasks
        super().__init__()

    def current_status(self):
        return print(f"Completed: {len(self.data)} of {self.total_tasks}", end="\r")


class JobsRunnerAsyncio:
    """A class for running a collection of interviews asynchronously.

    It gets instaniated from a Jobs object.
    The Jobs object is a collection of interviews that are to be run.
    """

    MAX_CONCURRENT_DEFAULT = 500

    def __init__(self, jobs: "Jobs"):
        self.jobs = jobs
        self.interviews: List["Interview"] = jobs.interviews()
        self.bucket_collection: "BucketCollection" = jobs.bucket_collection
        self.total_interviews: List["Interview"] = []
        self._initialized = threading.Event()

        from edsl.config import CONFIG

        self.MAX_CONCURRENT = int(CONFIG.get("EDSL_MAX_CONCURRENT_TASKS"))
        # print(f"MAX_CONCURRENT: {self.MAX_CONCURRENT}")

    # async def run_async_generator(
    #     self,
    #     cache: Cache,
    #     n: int = 1,
    #     stop_on_exception: bool = False,
    #     sidecar_model: Optional[LanguageModel] = None,
    #     total_interviews: Optional[List["Interview"]] = None,
    #     raise_validation_errors: bool = False,
    # ) -> AsyncGenerator["Result", None]:
    #     """Creates the tasks, runs them asynchronously, and returns the results as a Results object.

    #     Completed tasks are yielded as they are completed.

    #     :param n: how many times to run each interview
    #     :param stop_on_exception: Whether to stop the interview if an exception is raised
    #     :param sidecar_model: a language model to use in addition to the interview's model
    #     :param total_interviews: A list of interviews to run can be provided instead.
    #     :param raise_validation_errors: Whether to raise validation errors
    #     """
    #     tasks = []
    #     if total_interviews:  # was already passed in total interviews
    #         self.total_interviews = total_interviews
    #     else:
    #         self.total_interviews = list(
    #             self._populate_total_interviews(n=n)
    #         )  # Populate self.total_interviews before creating tasks
    #     self._initialized.set()  # Signal that we're ready

    #     for interview in self.total_interviews:
    #         interviewing_task = self._build_interview_task(
    #             interview=interview,
    #             stop_on_exception=stop_on_exception,
    #             sidecar_model=sidecar_model,
    #             raise_validation_errors=raise_validation_errors,
    #         )
    #         tasks.append(asyncio.create_task(interviewing_task))

    #     for task in asyncio.as_completed(tasks):
    #         result = await task
    #         self.jobs_runner_status.add_completed_interview(result)
    #         yield result

    async def run_async_generator(
        self,
        cache: Cache,
        n: int = 1,
        stop_on_exception: bool = False,
        sidecar_model: Optional[LanguageModel] = None,
        total_interviews: Optional[List["Interview"]] = None,
        raise_validation_errors: bool = False,
    ) -> AsyncGenerator["Result", None]:
        """Creates and processes tasks asynchronously, yielding results as they complete.

        Tasks are created and processed in a streaming fashion rather than building the full list upfront.
        Results are yielded as soon as they are available.

        :param n: how many times to run each interview
        :param stop_on_exception: Whether to stop the interview if an exception is raised
        :param sidecar_model: a language model to use in addition to the interview's model
        :param total_interviews: A list of interviews to run can be provided instead.
        :param raise_validation_errors: Whether to raise validation errors
        """
        # Initialize interviews iterator
        if total_interviews:
            interviews_iter = iter(total_interviews)
            self.total_interviews = total_interviews
        else:
            interviews_iter = self._populate_total_interviews(n=n)
            self.total_interviews = list(interviews_iter)
            interviews_iter = iter(self.total_interviews)  # Create fresh iterator

        self._initialized.set()  # Signal that we're ready

        # Keep track of active tasks
        active_tasks = set()

        try:
            while True:
                # Add new tasks if we're below max_concurrent and there are more interviews
                while len(active_tasks) < self.MAX_CONCURRENT:
                    try:
                        interview = next(interviews_iter)
                        task = asyncio.create_task(
                            self._build_interview_task(
                                interview=interview,
                                stop_on_exception=stop_on_exception,
                                sidecar_model=sidecar_model,
                                raise_validation_errors=raise_validation_errors,
                            )
                        )
                        active_tasks.add(task)
                        # Add callback to remove task from set when done
                        task.add_done_callback(active_tasks.discard)
                    except StopIteration:
                        break

                if not active_tasks:
                    break

                # Wait for next completed task
                done, _ = await asyncio.wait(
                    active_tasks, return_when=asyncio.FIRST_COMPLETED
                )

                # Process completed tasks
                for task in done:
                    try:
                        result = await task
                        self.jobs_runner_status.add_completed_interview(result)
                        yield result
                    except Exception as e:
                        if stop_on_exception:
                            # Cancel remaining tasks
                            for t in active_tasks:
                                if not t.done():
                                    t.cancel()
                            raise
                        else:
                            # Log error and continue
                            # logger.error(f"Task failed with error: {e}")
                            continue
        finally:
            # Ensure we cancel any remaining tasks if we exit early
            for task in active_tasks:
                if not task.done():
                    task.cancel()

    def _populate_total_interviews(
        self, n: int = 1
    ) -> Generator["Interview", None, None]:
        """Populates self.total_interviews with n copies of each interview.

        :param n: how many times to run each interview.
        """
        for interview in self.interviews:
            for iteration in range(n):
                if iteration > 0:
                    yield interview.duplicate(iteration=iteration, cache=self.cache)
                else:
                    interview.cache = self.cache
                    yield interview

    async def run_async(self, cache: Optional[Cache] = None, n: int = 1) -> Results:
        """Used for some other modules that have a non-standard way of running interviews."""
        self.jobs_runner_status = JobsRunnerStatus(self, n=n)
        self.cache = Cache() if cache is None else cache
        data = []
        async for result in self.run_async_generator(cache=self.cache, n=n):
            data.append(result)
        return Results(survey=self.jobs.survey, data=data)

    def simple_run(self):
        data = asyncio.run(self.run_async())
        return Results(survey=self.jobs.survey, data=data)

    async def _build_interview_task(
        self,
        *,
        interview: Interview,
        stop_on_exception: bool = False,
        sidecar_model: Optional["LanguageModel"] = None,
        raise_validation_errors: bool = False,
    ) -> "Result":
        """Conducts an interview and returns the result.

        :param interview: the interview to conduct
        :param stop_on_exception: stops the interview if an exception is raised
        :param sidecar_model: a language model to use in addition to the interview's model
        """
        # the model buckets are used to track usage rates
        model_buckets = self.bucket_collection[interview.model]

        # get the results of the interview
        answer, valid_results = await interview.async_conduct_interview(
            model_buckets=model_buckets,
            stop_on_exception=stop_on_exception,
            sidecar_model=sidecar_model,
            raise_validation_errors=raise_validation_errors,
        )

        question_results = {}
        for result in valid_results:
            question_results[result.question_name] = result

        answer_key_names = list(question_results.keys())

        generated_tokens_dict = {
            k + "_generated_tokens": question_results[k].generated_tokens
            for k in answer_key_names
        }
        comments_dict = {
            k + "_comment": question_results[k].comment for k in answer_key_names
        }

        # we should have a valid result for each question
        answer_dict = {k: answer[k] for k in answer_key_names}
        assert len(valid_results) == len(answer_key_names)

        # TODO: move this down into Interview
        question_name_to_prompts = dict({})
        for result in valid_results:
            question_name = result.question_name
            question_name_to_prompts[question_name] = {
                "user_prompt": result.prompts["user_prompt"],
                "system_prompt": result.prompts["system_prompt"],
            }

        prompt_dictionary = {}
        for answer_key_name in answer_key_names:
            prompt_dictionary[
                answer_key_name + "_user_prompt"
            ] = question_name_to_prompts[answer_key_name]["user_prompt"]
            prompt_dictionary[
                answer_key_name + "_system_prompt"
            ] = question_name_to_prompts[answer_key_name]["system_prompt"]

        raw_model_results_dictionary = {}
        cache_used_dictionary = {}
        for result in valid_results:
            question_name = result.question_name
            raw_model_results_dictionary[
                question_name + "_raw_model_response"
            ] = result.raw_model_response
            raw_model_results_dictionary[question_name + "_cost"] = result.cost
            one_use_buys = (
                "NA"
                if isinstance(result.cost, str)
                or result.cost == 0
                or result.cost is None
                else 1.0 / result.cost
            )
            raw_model_results_dictionary[question_name + "_one_usd_buys"] = one_use_buys
            cache_used_dictionary[question_name] = result.cache_used

        result = Result(
            agent=interview.agent,
            scenario=interview.scenario,
            model=interview.model,
            iteration=interview.iteration,
            answer=answer_dict,
            prompt=prompt_dictionary,
            raw_model_response=raw_model_results_dictionary,
            survey=interview.survey,
            generated_tokens=generated_tokens_dict,
            comments_dict=comments_dict,
            cache_used_dict=cache_used_dictionary,
        )
        result.interview_hash = hash(interview)

        return result

    @property
    def elapsed_time(self):
        return time.monotonic() - self.start_time

    def process_results(
        self, raw_results: Results, cache: Cache, print_exceptions: bool
    ):
        interview_lookup = {
            hash(interview): index
            for index, interview in enumerate(self.total_interviews)
        }
        interview_hashes = list(interview_lookup.keys())

        task_history = TaskHistory(self.total_interviews, include_traceback=False)

        results = Results(
            survey=self.jobs.survey,
            data=sorted(
                raw_results, key=lambda x: interview_hashes.index(x.interview_hash)
            ),
            task_history=task_history,
            cache=cache,
        )
        results.bucket_collection = self.bucket_collection

        if results.has_unfixed_exceptions and print_exceptions:
            from edsl.scenarios.FileStore import HTMLFileStore
            from edsl.config import CONFIG
            from edsl.coop.coop import Coop

            msg = f"Exceptions were raised in {len(results.task_history.indices)} out of {len(self.total_interviews)} interviews.\n"

            if len(results.task_history.indices) > 5:
                msg += f"Exceptions were raised in the following interviews: {results.task_history.indices}.\n"

            import sys

            print(msg, file=sys.stderr)
            from edsl.config import CONFIG

            if CONFIG.get("EDSL_OPEN_EXCEPTION_REPORT_URL") == "True":
                open_in_browser = True
            elif CONFIG.get("EDSL_OPEN_EXCEPTION_REPORT_URL") == "False":
                open_in_browser = False
            else:
                raise Exception(
                    "EDSL_OPEN_EXCEPTION_REPORT_URL", "must be either True or False"
                )

            # print("open_in_browser", open_in_browser)

            filepath = results.task_history.html(
                cta="Open report to see details.",
                open_in_browser=open_in_browser,
                return_link=True,
            )

            try:
                coop = Coop()
                user_edsl_settings = coop.edsl_settings
                remote_logging = user_edsl_settings["remote_logging"]
            except Exception as e:
                print(e)
                remote_logging = False

            if remote_logging:
                filestore = HTMLFileStore(filepath)
                coop_details = filestore.push(description="Error report")
                print(coop_details)

            print("Also see: https://docs.expectedparrot.com/en/latest/exceptions.html")

        return results

    @jupyter_nb_handler
    async def run(
        self,
        cache: Union[Cache, False, None],
        n: int = 1,
        stop_on_exception: bool = False,
        progress_bar: bool = False,
        sidecar_model: Optional[LanguageModel] = None,
        jobs_runner_status: Optional[Type[JobsRunnerStatusBase]] = None,
        job_uuid: Optional[UUID] = None,
        print_exceptions: bool = True,
        raise_validation_errors: bool = False,
    ) -> "Coroutine":
        """Runs a collection of interviews, handling both async and sync contexts."""

        self.results = []
        self.start_time = time.monotonic()
        self.completed = False
        self.cache = cache
        self.sidecar_model = sidecar_model

        from edsl.coop import Coop

        coop = Coop()
        endpoint_url = coop.get_progress_bar_url()

        if jobs_runner_status is not None:
            self.jobs_runner_status = jobs_runner_status(
                self, n=n, endpoint_url=endpoint_url, job_uuid=job_uuid
            )
        else:
            self.jobs_runner_status = JobsRunnerStatus(
                self, n=n, endpoint_url=endpoint_url, job_uuid=job_uuid
            )

        stop_event = threading.Event()

        async def process_results(cache):
            """Processes results from interviews."""
            async for result in self.run_async_generator(
                n=n,
                stop_on_exception=stop_on_exception,
                cache=cache,
                sidecar_model=sidecar_model,
                raise_validation_errors=raise_validation_errors,
            ):
                self.results.append(result)
            self.completed = True

        def run_progress_bar(stop_event):
            """Runs the progress bar in a separate thread."""
            self.jobs_runner_status.update_progress(stop_event)

        if progress_bar and self.jobs_runner_status.has_ep_api_key():
            self.jobs_runner_status.setup()
            progress_thread = threading.Thread(
                target=run_progress_bar, args=(stop_event,)
            )
            progress_thread.start()
        elif progress_bar:
            warnings.warn(
                "You need an Expected Parrot API key to view job progress bars."
            )

        exception_to_raise = None
        try:
            with cache as c:
                await process_results(cache=c)
        except KeyboardInterrupt:
            print("Keyboard interrupt received. Stopping gracefully...")
            stop_event.set()
        except Exception as e:
            if stop_on_exception:
                exception_to_raise = e
            stop_event.set()
        finally:
            stop_event.set()
            if progress_bar and self.jobs_runner_status.has_ep_api_key():
                # self.jobs_runner_status.stop_event.set()
                if progress_thread:
                    progress_thread.join()

            if exception_to_raise:
                raise exception_to_raise

            return self.process_results(
                raw_results=self.results, cache=cache, print_exceptions=print_exceptions
            )



from typing import Optional, Union, Literal
import requests
import sys
from edsl.exceptions.coop import CoopServerResponseError

# from edsl.enums import VisibilityType
from edsl.results import Results


class JobsRemoteInferenceHandler:
    def __init__(self, jobs, verbose=False, poll_interval=3):
        """
        >>> from edsl.jobs import Jobs
        >>> jh = JobsRemoteInferenceHandler(Jobs.example(), verbose=True)
        >>> jh.use_remote_inference(True)
        False
        >>> jh._poll_remote_inference_job({'uuid':1234}, testing_simulated_response={"status": "failed"}) # doctest: +NORMALIZE_WHITESPACE
        Job failed.
        ...
        >>> jh._poll_remote_inference_job({'uuid':1234}, testing_simulated_response={"status": "completed"}) # doctest: +NORMALIZE_WHITESPACE
        Job completed and Results stored on Coop: None.
        Results(...)
        """
        self.jobs = jobs
        self.verbose = verbose
        self.poll_interval = poll_interval

        self._remote_job_creation_data = None
        self._job_uuid = None

    @property
    def remote_job_creation_data(self):
        return self._remote_job_creation_data

    @property
    def job_uuid(self):
        return self._job_uuid

    def use_remote_inference(self, disable_remote_inference: bool) -> bool:
        if disable_remote_inference:
            return False
        if not disable_remote_inference:
            try:
                from edsl import Coop

                user_edsl_settings = Coop().edsl_settings
                return user_edsl_settings.get("remote_inference", False)
            except requests.ConnectionError:
                pass
            except CoopServerResponseError as e:
                pass

        return False

    def create_remote_inference_job(
        self,
        iterations: int = 1,
        remote_inference_description: Optional[str] = None,
        remote_inference_results_visibility: Optional["VisibilityType"] = "unlisted",
        verbose=False,
    ):
        """ """
        from edsl.config import CONFIG
        from edsl.coop.coop import Coop
        from rich import print as rich_print

        coop = Coop()
        print("Remote inference activated. Sending job to server...")
        remote_job_creation_data = coop.remote_inference_create(
            self.jobs,
            description=remote_inference_description,
            status="queued",
            iterations=iterations,
            initial_results_visibility=remote_inference_results_visibility,
        )
        job_uuid = remote_job_creation_data.get("uuid")
        print(f"Job sent to server. (Job uuid={job_uuid}).")

        expected_parrot_url = CONFIG.get("EXPECTED_PARROT_URL")
        progress_bar_url = f"{expected_parrot_url}/home/remote-job-progress/{job_uuid}"

        rich_print(
            f"View job progress here: [#38bdf8][link={progress_bar_url}]{progress_bar_url}[/link][/#38bdf8]"
        )

        self._remote_job_creation_data = remote_job_creation_data
        self._job_uuid = job_uuid
        # return remote_job_creation_data

    @staticmethod
    def check_status(job_uuid):
        from edsl.coop.coop import Coop

        coop = Coop()
        return coop.remote_inference_get(job_uuid)

    def poll_remote_inference_job(self):
        return self._poll_remote_inference_job(
            self.remote_job_creation_data, verbose=self.verbose
        )

    def _poll_remote_inference_job(
        self,
        remote_job_creation_data: dict,
        verbose=False,
        poll_interval: Optional[float] = None,
        testing_simulated_response: Optional[dict] = None,
    ) -> Union[Results, None]:
        import time
        from datetime import datetime
        from edsl.config import CONFIG
        from edsl.coop.coop import Coop

        if poll_interval is None:
            poll_interval = self.poll_interval

        expected_parrot_url = CONFIG.get("EXPECTED_PARROT_URL")

        job_uuid = remote_job_creation_data.get("uuid")
        coop = Coop()

        if testing_simulated_response is not None:
            remote_job_data_fetcher = lambda job_uuid: testing_simulated_response
            object_fetcher = (
                lambda results_uuid, expected_object_type: Results.example()
            )
        else:
            remote_job_data_fetcher = coop.remote_inference_get
            object_fetcher = coop.get

        job_in_queue = True
        while job_in_queue:
            remote_job_data = remote_job_data_fetcher(job_uuid)
            status = remote_job_data.get("status")
            if status == "cancelled":
                print("\r" + " " * 80 + "\r", end="")
                print("Job cancelled by the user.")
                print(
                    f"See {expected_parrot_url}/home/remote-inference for more details."
                )
                return None
            elif status == "failed":
                print("\r" + " " * 80 + "\r", end="")
                # write to stderr
                latest_error_report_url = remote_job_data.get("latest_error_report_url")
                if latest_error_report_url:
                    print("Job failed.")
                    print(
                        f"Your job generated exceptions. Details on these exceptions can be found in the following report: {latest_error_report_url}"
                    )
                    print(
                        f"Need support? Post a message at the Expected Parrot Discord channel (https://discord.com/invite/mxAYkjfy9m) or send an email to info@expectedparrot.com."
                    )
                else:
                    print("Job failed.")
                    print(
                        f"See {expected_parrot_url}/home/remote-inference for more details."
                    )
                return None
            elif status == "completed":
                results_uuid = remote_job_data.get("results_uuid")
                results_url = remote_job_data.get("results_url")
                results = object_fetcher(results_uuid, expected_object_type="results")
                print("\r" + " " * 80 + "\r", end="")
                print(f"Job completed and Results stored on Coop: {results_url}.")
                return results
            else:
                duration = poll_interval
                time_checked = datetime.now().strftime("%Y-%m-%d %I:%M:%S %p")
                frames = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
                start_time = time.time()
                i = 0
                while time.time() - start_time < duration:
                    print(
                        f"\r{frames[i % len(frames)]} Job status: {status} - last update: {time_checked}",
                        end="",
                        flush=True,
                    )
                    time.sleep(0.1)
                    i += 1

    def use_remote_inference(self, disable_remote_inference: bool) -> bool:
        if disable_remote_inference:
            return False
        if not disable_remote_inference:
            try:
                from edsl import Coop

                user_edsl_settings = Coop().edsl_settings
                return user_edsl_settings.get("remote_inference", False)
            except requests.ConnectionError:
                pass
            except CoopServerResponseError as e:
                pass

        return False

    async def create_and_poll_remote_job(
        self,
        iterations: int = 1,
        remote_inference_description: Optional[str] = None,
        remote_inference_results_visibility: Optional[
            Literal["private", "public", "unlisted"]
        ] = "unlisted",
    ) -> Union[Results, None]:
        """
        Creates and polls a remote inference job asynchronously.
        Reuses existing synchronous methods but runs them in an async context.

        :param iterations: Number of times to run each interview
        :param remote_inference_description: Optional description for the remote job
        :param remote_inference_results_visibility: Visibility setting for results
        :return: Results object if successful, None if job fails or is cancelled
        """
        import asyncio
        from functools import partial

        # Create job using existing method
        loop = asyncio.get_event_loop()
        remote_job_creation_data = await loop.run_in_executor(
            None,
            partial(
                self.create_remote_inference_job,
                iterations=iterations,
                remote_inference_description=remote_inference_description,
                remote_inference_results_visibility=remote_inference_results_visibility,
            ),
        )

        # Poll using existing method but with async sleep
        return await loop.run_in_executor(
            None, partial(self.poll_remote_inference_job, remote_job_creation_data)
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl.inference_services.InferenceServiceABC import InferenceServiceABC
import warnings


class InferenceServicesCollection:
    added_models = {}

    def __init__(self, services: list[InferenceServiceABC] = None):
        self.services = services or []

    @classmethod
    def add_model(cls, service_name, model_name):
        if service_name not in cls.added_models:
            cls.added_models[service_name] = []
        cls.added_models[service_name].append(model_name)

    @staticmethod
    def _get_service_available(service, warn: bool = False) -> list[str]:
        try:
            service_models = service.available()
        except Exception:
            if warn:
                warnings.warn(
                    f"""Error getting models for {service._inference_service_}. 
                    Check that you have properly stored your Expected Parrot API key and activated remote inference, or stored your own API keys for the language models that you want to use.
                    See https://docs.expectedparrot.com/en/latest/api_keys.html for instructions on storing API keys.
                    Relying on Coop.""",
                    UserWarning,
                )

            # Use the list of models on Coop as a fallback
            try:
                from edsl import Coop

                c = Coop()
                models_from_coop = c.fetch_models()
                service_models = models_from_coop.get(service._inference_service_, [])

                # cache results
                service._models_list_cache = service_models

            # Finally, use the available models cache from the Python file
            except Exception:
                if warn:
                    warnings.warn(
                        f"""Error getting models for {service._inference_service_}. 
                        Relying on EDSL cache.""",
                        UserWarning,
                    )

                from edsl.inference_services.models_available_cache import (
                    models_available,
                )

                service_models = models_available.get(service._inference_service_, [])

                # cache results
                service._models_list_cache = service_models

        return service_models

    def available(self):
        total_models = []
        for service in self.services:
            service_models = self._get_service_available(service)
            for model in service_models:
                total_models.append([model, service._inference_service_, -1])

            for model in self.added_models.get(service._inference_service_, []):
                total_models.append([model, service._inference_service_, -1])

        sorted_models = sorted(total_models)
        for i, model in enumerate(sorted_models):
            model[2] = i
            model = tuple(model)
        return sorted_models

    def register(self, service):
        self.services.append(service)

    def create_model_factory(self, model_name: str, service_name=None, index=None):
        from edsl.inference_services.TestService import TestService

        if model_name == "test":
            return TestService.create_model(model_name)

        if service_name:
            for service in self.services:
                if service_name == service._inference_service_:
                    return service.create_model(model_name)

        for service in self.services:
            if model_name in self._get_service_available(service):
                if service_name is None or service_name == service._inference_service_:
                    return service.create_model(model_name)

        raise Exception(f"Model {model_name} not found in any of the services")



import os
from edsl.exceptions import MissingAPIKeyError


class JobsChecks:
    def __init__(self, jobs):
        """ """
        self.jobs = jobs

    def check_api_keys(self) -> None:
        from edsl import Model

        for model in self.jobs.models + [Model()]:
            if not model.has_valid_api_key():
                raise MissingAPIKeyError(
                    model_name=str(model.model),
                    inference_service=model._inference_service_,
                )

    def get_missing_api_keys(self) -> set:
        """
        Returns a list of the api keys that a user needs to run this job, but does not currently have in their .env file.
        """
        missing_api_keys = set()

        from edsl import Model
        from edsl.enums import service_to_api_keyname

        for model in self.jobs.models + [Model()]:
            if not model.has_valid_api_key():
                key_name = service_to_api_keyname.get(
                    model._inference_service_, "NOT FOUND"
                )
                missing_api_keys.add(key_name)

        return missing_api_keys

    def user_has_ep_api_key(self) -> bool:
        """
        Returns True if the user has an EXPECTED_PARROT_API_KEY in their env.

        Otherwise, returns False.
        """

        coop_api_key = os.getenv("EXPECTED_PARROT_API_KEY")

        if coop_api_key is not None:
            return True
        else:
            return False

    def user_has_all_model_keys(self):
        """
        Returns True if the user has all model keys required to run their job.

        Otherwise, returns False.
        """

        try:
            self.check_api_keys()
            return True
        except MissingAPIKeyError:
            return False
        except Exception:
            raise

    def needs_external_llms(self) -> bool:
        """
        Returns True if the job needs external LLMs to run.

        Otherwise, returns False.
        """
        # These cases are necessary to skip the API key check during doctests

        # Accounts for Results.example()
        all_agents_answer_questions_directly = len(self.jobs.agents) > 0 and all(
            [hasattr(a, "answer_question_directly") for a in self.jobs.agents]
        )

        # Accounts for InterviewExceptionEntry.example()
        only_model_is_test = set([m.model for m in self.jobs.models]) == set(["test"])

        # Accounts for Survey.__call__
        all_questions_are_functional = set(
            [q.question_type for q in self.jobs.survey.questions]
        ) == set(["functional"])

        if (
            all_agents_answer_questions_directly
            or only_model_is_test
            or all_questions_are_functional
        ):
            return False
        else:
            return True

    def needs_key_process(self):
        return (
            not self.user_has_all_model_keys()
            and not self.user_has_ep_api_key()
            and self.needs_external_llms()
        )

    def key_process(self):
        import secrets
        from dotenv import load_dotenv
        from edsl import CONFIG
        from edsl.coop.coop import Coop
        from edsl.utilities.utilities import write_api_key_to_env

        missing_api_keys = self.get_missing_api_keys()

        edsl_auth_token = secrets.token_urlsafe(16)

        print("You're missing some of the API keys needed to run this job:")
        for api_key in missing_api_keys:
            print(f"     🔑 {api_key}")
        print(
            "\nYou can either add the missing keys to your .env file, or use remote inference."
        )
        print("Remote inference allows you to run jobs on our server.")
        print("\n🚀 To use remote inference, sign up at the following link:")

        coop = Coop()
        coop._display_login_url(edsl_auth_token=edsl_auth_token)

        print(
            "\nOnce you log in, we will automatically retrieve your Expected Parrot API key and continue your job remotely."
        )

        api_key = coop._poll_for_api_key(edsl_auth_token)

        if api_key is None:
            print("\nTimed out waiting for login. Please try again.")
            return

        write_api_key_to_env(api_key)
        print("✨ API key retrieved and written to .env file.\n")

        # Retrieve API key so we can continue running the job
        load_dotenv()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from tabulate import tabulate
from pathlib import Path

from edsl.results.CSSParameterizer import CSSParameterizer


class TableDisplay:
    max_height = 400
    min_height = 200

    @classmethod
    def get_css(cls):
        """Load CSS content from the file next to this module"""
        css_path = Path(__file__).parent / "table_display.css"
        return css_path.read_text()

    def __init__(self, headers, data, tablefmt=None, raw_data_set=None):
        self.headers = headers
        self.data = data
        self.tablefmt = tablefmt
        self.raw_data_set = raw_data_set

        if hasattr(raw_data_set, "print_parameters"):
            if raw_data_set.print_parameters:
                self.printing_parameters = raw_data_set.print_parameters
            else:
                self.printing_parameters = {}
        else:
            self.printing_parameters = {}

    def to_csv(self, filename: str):
        self.raw_data_set.to_csv(filename)

    def write(self, filename: str):
        if self.tablefmt is None:
            table = tabulate(self.data, headers=self.headers, tablefmt="simple")
        else:
            table = tabulate(self.data, headers=self.headers, tablefmt=self.tablefmt)

        with open(filename, "w") as file:
            print("Writing table to", filename)
            file.write(table)

    def to_pandas(self):
        return self.raw_data_set.to_pandas()

    def to_list(self):
        return self.raw_data_set.to_list()

    def __repr__(self):
        from tabulate import tabulate

        if self.tablefmt is None:
            return tabulate(self.data, headers=self.headers, tablefmt="simple")
        else:
            return tabulate(self.data, headers=self.headers, tablefmt=self.tablefmt)

    def long(self):
        new_header = ["row", "key", "value"]
        new_data = []
        for index, row in enumerate(self.data):
            new_data.extend([[index, k, v] for k, v in zip(self.headers, row)])
        return TableDisplay(new_header, new_data)

    def _repr_html_(self):
        if self.tablefmt is not None:
            return (
                "<pre>"
                + tabulate(self.data, headers=self.headers, tablefmt=self.tablefmt)
                + "</pre>"
            )

        num_rows = len(self.data)
        height = min(
            num_rows * 30 + 50, self.max_height
        )  # Added extra space for header

        if height < self.min_height:
            height = self.min_height

        html_template = """
        <style>
            {css}
        </style>
        <div class="table-container">
            <div class="scroll-table-wrapper">
                {table}
            </div>
        </div>
        """

        html_content = tabulate(self.data, headers=self.headers, tablefmt="html")
        html_content = html_content.replace("<table>", '<table class="scroll-table">')

        height_string = f"{height}px"
        parameters = {"containerHeight": height_string, "headerColor": "blue"}
        parameters.update(self.printing_parameters)
        rendered_css = CSSParameterizer(self.get_css()).apply_parameters(parameters)

        return html_template.format(table=html_content, css=rendered_css)

    @classmethod
    def example(
        cls,
        headers=None,
        data=None,
        filename: str = "table_example.html",
        auto_open: bool = True,
    ):
        """
        Creates a standalone HTML file with an example table in an iframe and optionally opens it in a new tab.

        Args:
            cls: The class itself
            headers (list): List of column headers. If None, uses example headers
            data (list): List of data rows. If None, uses example data
            filename (str): The name of the HTML file to create. Defaults to "table_example.html"
            auto_open (bool): Whether to automatically open the file in the default web browser. Defaults to True

        Returns:
            str: The path to the created HTML file
        """
        import os
        import webbrowser

        # Use example data if none provided
        if headers is None:
            headers = ["Name", "Age", "City", "Occupation"]
        if data is None:
            data = [
                [
                    "John Doe",
                    30,
                    "New York",
                    """cls: The class itself
        headers (list): List of column headers. If None, uses example headers
        data (list): List of data rows. If None, uses example data
        filename (str): The name of the HTML file to create. Defaults to "table_example.html"
        auto_open (bool): Whether to automatically open the file in the default web browser. Defaults to True
        """,
                ],
                ["Jane Smith", 28, "San Francisco", "Designer"],
                ["Bob Johnson", 35, "Chicago", "Manager"],
                ["Alice Brown", 25, "Boston", "Developer"],
                ["Charlie Wilson", 40, "Seattle", "Architect"],
            ]

        # Create instance with the data
        instance = cls(headers=headers, data=data)

        # Get the table HTML content
        table_html = instance._repr_html_()

        # Calculate the appropriate iframe height
        num_rows = len(data)
        iframe_height = min(num_rows * 140 + 50, cls.max_height)
        print(f"Table height: {iframe_height}px")

        # Create the full HTML document
        html_content = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>Table Display Example</title>
            <style>
                body {{
                    margin: 0;
                    padding: 20px;
                    font-family: Arial, sans-serif;
                }}
                iframe {{
                    width: 100%;
                    height: {iframe_height}px;
                    border: none;
                    overflow: hidden;
                }}
            </style>
        </head>
        <body>
            <iframe srcdoc='{table_html}'></iframe>
        </body>
        </html>
        """

        # Write the HTML file
        abs_path = os.path.abspath(filename)
        with open(filename, "w", encoding="utf-8") as f:
            f.write(html_content)

        # Open in browser if requested
        if auto_open:
            webbrowser.open("file://" + abs_path, new=2)

        return abs_path


if __name__ == "__main__":
    TableDisplay.example()



"""A module for displaying data in various formats."""

from html import escape


def create_image(console, image_filename):
    """Create an image from the console output."""
    font_size = 15
    from PIL import Image, ImageDraw, ImageFont

    text = console.export_text()  # Get the console output as text.

    # Create an image from the text
    font_size = 15
    font = ImageFont.load_default()  # Use the default font to avoid file path issues.
    # text_width, text_height = ImageDraw.Draw(
    #    Image.new("RGB", (100, 100))
    # ).multiline_textsize(text, font=font)
    text_width, text_height = get_multiline_textsize(text, font)
    image = Image.new(
        "RGB", (text_width + 20, text_height + 20), color=(255, 255, 255)
    )  # Add some padding
    d = ImageDraw.Draw(image)

    # Draw text to image
    d.multiline_text((10, 10), text, font=font, fill=(0, 0, 0))
    # Save the image
    image.save(image_filename)


def display_table(console, table, filename):
    # from rich.console import Console
    # from rich.table import Table
    """Display the table using the rich library and save it to a file if a filename is provided."""
    if filename is not None:
        with open(filename, "w") as f:
            with console.capture() as capture:
                console.print(table)
            f.write(capture.get())
        create_image(console, filename + ".png")
    else:
        console.print(table)


def gen_html_sandwich(html_inner, interactive=False):
    """Wrap the inner HTML content in a header and footer to make a complete HTML document."""
    return html_inner
    if interactive:
        html_header = """
            <html>
            <head>
            <script src="https://ajax.googleapis.com/ajax/libs/jquery/3.7.1/jquery.min.js"></script>
            <link rel="stylesheet" href="https://cdn.datatables.net/1.13.6/css/jquery.dataTables.css" />
            <script src="https://cdn.datatables.net/1.13.6/js/jquery.dataTables.js"></script>
            <style>
                table {
                    font-family: Arial, sans-serif;
                    border-collapse: collapse;
                    width: 100%;
                }
                
                td, th {
                    border: 1px solid #dddddd;
                    text-align: left;
                    padding: 8px;
                }
                
                tr:nth-child(even) {
                    background-color: #dddddd;
                }
            </style>
            <script>
            $(document).ready( function () {
                $('#myTable').DataTable();
            } )
            </script>
            </head>
            <body>
        """
    else:
        html_header = """
            <html>
            <head>
            <style>
                table {
                    font-family: Arial, sans-serif;
                    border-collapse: collapse;
                    width: 100%;
                }
                
                td, th {
                    border: 1px solid #dddddd;
                    text-align: left;
                    padding: 8px;
                }
                
                tr:nth-child(even) {
                    background-color: #dddddd;
                }
            </style>
            </head>
            <body>
        """

    html_footer = """
        </body>
        </html>
    """
    return html_header + html_inner + html_footer


def view_html(html):
    """Display HTML content in a web browser."""
    import tempfile
    import webbrowser

    with tempfile.NamedTemporaryFile("w", delete=False, suffix=".html") as f:
        url = "file://" + f.name
        # Write the HTML content to the file
        f.write(html)

    # Open the URL in the web browser
    webbrowser.open(url)


def human_readable_labeler_creator():
    """Create a function that maps thread ids to human-readable labels.

    It is structured as a closure, so that the mapping is persistent.
    I.e., when the returned function is called, it will use the same
    dictionary to map thread ids to human-readable labels if it's seen that ID
    before; otherwise, it will add a new entry to the dictionary.
    This will persist across calls to the function.
    """
    d = {}

    def func(thread_id):
        if thread_id in d:
            return d[thread_id]
        else:
            d[thread_id] = len(d)
            return d[thread_id]

    return func


def get_multiline_textsize(text, font):
    """Get the size of the text when it is drawn on an image."""
    lines = text.split("\n")

    # Initialize width and height
    max_width = 0
    total_height = 0

    for line in lines:
        # Get the size of the text for the line
        box = font.getbbox(line)
        width, height = box[2], box[3]

        # Update max_width if width of the current line is greater than max_width
        max_width = max(max_width, width)

        # Add height to total_height
        total_height += height

    return max_width, total_height


def print_results_long(results, max_rows=None):
    from rich.console import Console
    from rich.table import Table

    console = Console(record=True)
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Result index", style="dim")
    table.add_column("Key", style="dim")
    table.add_column("Value", style="dim")
    list_of_dicts = results.to_dicts()
    num_rows = 0
    for i, results_dict in enumerate(list_of_dicts):
        for key, value in results_dict.items():
            table.add_row(str(i), key, str(value))
            num_rows += 1
        if max_rows is not None and num_rows >= max_rows:
            break
    console.print(table)


def print_dict_with_rich(d, key_name="Key", value_name="Value", filename=None):
    """Print a dictionary as a table using the rich library.

    Example:
    >>> print_dict_with_rich({"a": 1, "b": 2, "c": 3})
    ┏━━━━━┳━━━━━━━┓
    ┃ Key ┃ Value ┃
    ┡━━━━━╇━━━━━━━┩
    │ a   │ 1     │
    │ b   │ 2     │
    │ c   │ 3     │
    └─────┴───────┘
    """
    from rich.console import Console
    from rich.table import Table

    console = Console(record=True)
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column(key_name, style="dim")
    table.add_column(value_name, style="dim")
    for key, value in d.items():
        table.add_row(key, str(value))
    console.print(table)
    # display_table(console, table, filename)


def print_dict_as_html_table(
    d,
    show=False,
    key_name="Key",
    value_name="Value",
    filename=None,
):
    """Print a dictionary as an HTML table.

    :param d: The dictionary to print.
    :param show: Whether to display the HTML table in the browser.
    :param key_name: The name of the key column.
    :param value_name: The name of the value column.
    :param filename: The name of the file to save the HTML table to.
    """
    # Start the HTML table
    html_table = f'<table border="1">\n<tr><th>{escape(key_name)}</th><th>{escape(value_name)}</th></tr>\n'

    # Add rows to the HTML table
    for key, value in d.items():
        html_table += (
            f"<tr><td>{escape(str(key))}</td><td>{escape(str(value))}</td></tr>\n"
        )

    # Close the HTML table
    html_table += "</table>"

    # Print the HTML table to console
    # print(html_table)

    # Write to file if a filename is provided
    if filename:
        with open(filename, "w") as file:
            file.write(html_table)
    else:
        if show:
            view_html(gen_html_sandwich(html_table))
        else:
            return html_table


def print_scenario_list(data):
    from rich.console import Console
    from rich.table import Table

    new_data = []
    for obs in data:
        try:
            _ = obs.pop("edsl_version")
            _ = obs.pop("edsl_class_name")
        except KeyError as e:
            # print(e)
            pass
        new_data.append(obs)

    columns = list(new_data[0].keys())
    console = Console(record=True)

    # Create a table object
    table = Table(show_header=True, header_style="bold magenta", show_lines=True)
    for column in columns:
        table.add_column(column, style="dim")

    for obs in new_data:
        row = [str(obs[key]) for key in columns]
        table.add_row(*row)

    console.print(table)


def print_list_of_dicts_with_rich(data, filename=None, split_at_dot=True):
    raise Exception(
        "print_list_of_dicts_with_rich is now called print_dataset_with_rich"
    )


def print_dataset_with_rich(data, filename=None, split_at_dot=True):
    """Initialize console object."""
    """
    The list seems superfluous.
    This prints a list of dictionaries as a table using the rich library.

    >>> data = [{"a": [1, 2, 3], "b": [4, 5, 6]}]
    >>> print_list_of_dicts_with_rich(data)
    ┏━━━┳━━━┓
    ┃ a ┃ b ┃
    ┡━━━╇━━━┩
    │ 1 │ 4 │
    ├───┼───┤
    │ 2 │ 5 │
    ├───┼───┤
    │ 3 │ 6 │
    └───┴───┘
    """
    from rich.console import Console
    from rich.table import Table

    console = Console(record=True)

    # Create a table object
    table = Table(show_header=True, header_style="bold magenta", show_lines=True)

    # Adding columns to the table
    for d in data:
        for key in d.keys():
            if split_at_dot:
                value = key.replace(".", "\n.")
            else:
                value = key
            table.add_column(value, style="dim")

    # Adding rows to the table
    num_rows = len(next(iter(data[0].values())))
    for i in range(num_rows):
        row = [str(d[key][i]) for d in data for key in d.keys()]
        table.add_row(*row)

    console.print(table)
    # display_table(console, table, filename)


def create_latex_table_from_data(data, filename=None, split_at_dot=True):
    """
    This function takes a list of dictionaries and returns a LaTeX table as a string.
    The table can either be printed or written to a file.

    >>> data = [{"a": [1, 2, 3], "b": [4, 5, 6]}]
    >>> print(create_latex_table_from_data(data))
    \\begin{tabular}{|c|c|}
    \\hline
    a & b \\\\
    \\hline
    1 & 4 \\\\
    2 & 5 \\\\
    3 & 6 \\\\
    \\hline
    \\end{tabular}
    """

    def escape_latex(s):
        replacements = [
            ("_", r"\_"),
            ("&", r"\&"),
            ("%", r"\%"),
            ("$", r"\$"),
            ("#", r"\#"),
            ("{", r"\{"),
            ("}", r"\}"),
            ("~", r"\textasciitilde{}"),
            ("^", r"\textasciicircum{}"),
            ("\\", r"\textbackslash{}"),
        ]

        for old, new in replacements:
            s = s.replace(old, new)
        return s

    # Start the LaTeX table
    latex_table = ["\\begin{tabular}{|" + "c|" * len(data[0]) + "}"]
    latex_table.append("\\hline")

    # Add the header row
    headers = []
    for key in data[0].keys():
        if split_at_dot:
            value = key.replace(".", "\n.")
        else:
            value = key
        headers.append(escape_latex(value))
    latex_table.append(" & ".join(headers) + " \\\\")
    latex_table.append("\\hline")

    # Determine the number of rows
    num_rows = len(next(iter(data[0].values())))

    # Debugging: Print the keys of the dictionaries
    # print("Keys in data[0]:", list(data[0].keys()))

    # Add the data rows
    for i in range(num_rows):
        row = []
        for key in data[0].keys():
            for d in data:
                try:
                    row.append(escape_latex(str(d[key][i])))
                except KeyError as e:
                    print(
                        f"KeyError: {e} - Key '{key}' not found in data dictionary. The keys are {list(d.keys())}"
                    )
                    raise
        latex_table.append(" & ".join(row) + " \\\\")

    latex_table.append("\\hline")
    latex_table.append("\\end{tabular}")

    # Join all parts into a single string
    latex_table_str = "\n".join(latex_table)

    # Write to file if filename is provided
    if filename:
        with open(filename, "w") as f:
            f.write(latex_table_str)
            print(f"Table written to {filename}")

    return latex_table_str


def print_list_of_dicts_as_html_table(data, interactive=True):
    """Print a list of dictionaries as an HTML table.

    :param data: The list of dictionaries to print.
    :param filename: The name of the file to save the HTML table to.
    :param interactive: Whether to make the table interactive using DataTables.
    """
    style = """
    <style>
        table {
            width: 100%;
            border-collapse: collapse;
        }
        table, th, td {
            border: 1px solid black;
        }
        th, td {
            padding: 10px;
            text-align: left;
        }
    </style>
    """
    html_table = style + '<table id="myTable" class="display">\n'
    html_table += "  <thead>\n"
    # Add the header row
    headers = [key for d in data for key in d.keys()]
    html_table += "  <tr>\n"
    for header in headers:
        html_table += f"    <th>{header}</th>\n"
    html_table += "  </tr>\n"
    html_table += "  </thead>\n</tbody>\n"

    # Determine the number of rows
    num_rows = max(len(values) for d in data for values in d.values())

    # Add the data rows
    for i in range(num_rows):
        html_table += "  <tr>\n"
        for d in data:
            for key in d.keys():
                value = d[key][i] if i < len(d[key]) else ""
                html_table += f"    <td>{value}</td>\n"
        html_table += "  </tr>\n"

    # Close the table
    html_table += "</tbody>\n"
    html_table += "</table>"
    return gen_html_sandwich(html_table, interactive=interactive)


def print_list_of_dicts_as_markdown_table(data, filename=None):
    """Print a list of dictionaries as a Markdown table.

    :param data: The list of dictionaries to print.
    :param filename: The name of the file to save the Markdown table to.
    """
    if not data:
        print("No data provided")
        return

    # Gather all unique headers
    # headers = list({key for d in data for key in d.keys()})
    headers = []
    for column in data:
        headers.append(list(column.keys())[0])

    markdown_table = "| " + " | ".join(headers) + " |\n"
    markdown_table += "|-" + "-|-".join(["" for _ in headers]) + "-|\n"

    num_rows = len(next(iter(data[0].values())))
    for i in range(num_rows):
        row = [str(d[key][i]) for d in data for key in d.keys()]
        # table.add_row(*row)
        markdown_table += "| " + " | ".join(row) + " |\n"

    # Output or save to file
    if filename:
        with open(filename, "w") as f:
            f.write(markdown_table)
    else:
        print(markdown_table)


def print_public_methods_with_doc(obj):
    """Print the public methods of an object along with their docstrings."""
    from rich.console import Console
    from rich.table import Table

    console = Console()
    public_methods_with_docstrings = [
        (method, getattr(obj, method).__doc__)
        for method in dir(obj)
        if callable(getattr(obj, method))
        and not method.startswith("_")
        and method != "methods"
    ]

    for method, doc in public_methods_with_docstrings:
        if doc:
            console.print(f"[bold]{method}:[/bold]", style="green")
            console.print(f"\t{doc.strip()}", style="yellow")


def print_tally_with_rich(data, filename=None):
    """Print a tally of values in a list using the rich library.

    Example:
    >>> data = {'a':12, 'b':14, 'c':9}
    >>> print_tally_with_rich(data)
    ┏━━━━━━━┳━━━━━━━┓
    ┃ Value ┃ Count ┃
    ┡━━━━━━━╇━━━━━━━┩
    │ a     │ 12    │
    │ b     │ 14    │
    │ c     │ 9     │
    └───────┴───────┘
    """
    # Initialize a console object
    from rich.console import Console
    from rich.table import Table
    from IPython.display import display

    console = Console(record=True)

    # Create a new table
    table = Table(show_header=True, header_style="bold magenta", row_styles=["", "dim"])

    # Add columns to the table
    table.add_column("Value", style="dim")
    table.add_column("Count", style="dim")

    # Add rows to the table
    for key, value in data.items():
        table.add_row(key, str(value))

    from IPython.display import display

    display_table(console, table, filename)


def print_table_with_rich(data, filename=None):
    """Print a list of dictionaries as a table using the rich library.

    Example:
    >>> data = [{"a": 1, "b": 2, "c": 3}]
    >>> print_table_with_rich(data)
    ┏━━━┳━━━┳━━━┓
    ┃ a ┃ b ┃ c ┃
    ┡━━━╇━━━╇━━━┩
    │ 1 │ 2 │ 3 │
    └───┴───┴───┘
    >>> data = [{"a": 1, "b": 2, "c": 3},{"a": 2, "b": 9, "c": 8}]
    >>> print_table_with_rich(data)
    ┏━━━┳━━━┳━━━┓
    ┃ a ┃ b ┃ c ┃
    ┡━━━╇━━━╇━━━┩
    │ 1 │ 2 │ 3 │
    │ 2 │ 9 │ 8 │
    └───┴───┴───┘
    """
    from rich.console import Console
    from rich.table import Table

    # Initialize a console object - expects a list of dictionaries
    console = Console(record=True)

    # Create a new table
    table = Table(show_header=True, header_style="bold magenta", row_styles=["", "dim"])

    # Check if data is empty; if it is, exit
    if not data:
        console.print("No data provided!")
        return

    # Add columns based on keys in the first dictionary
    for key in data[0].keys():
        table.add_column(key, style="dim")

    # Add rows to the table
    for row in data:
        table.add_row(*[str(value) for value in row.values()])

    display_table(console, table, filename)


if __name__ == "__main__":
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3})
    import doctest

    doctest.testmod()
    # print_list_of_dicts_with_rich([{"a": [1, 2, 3], "b": [4, 5, 6]}])
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html")
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c



"""A list of Agent objects.

Example usage:

.. code-block:: python

    al = AgentList([Agent.example(), Agent.example()])
    len(al)
    2
    
"""

from __future__ import annotations
import csv
import json
from collections import UserList
from typing import Any, List, Optional, Union, TYPE_CHECKING
from rich import print_json
from rich.table import Table
from simpleeval import EvalWithCompoundTypes
from edsl.Base import Base
from edsl.utilities.decorators import add_edsl_version, remove_edsl_version

from collections.abc import Iterable

from edsl.exceptions.agents import AgentListError

if TYPE_CHECKING:
    from edsl.scenarios.ScenarioList import ScenarioList


def is_iterable(obj):
    return isinstance(obj, Iterable)


class AgentList(UserList, Base):
    """A list of Agents."""

    __documentation__ = (
        "https://docs.expectedparrot.com/en/latest/agents.html#agentlist-class"
    )

    def __init__(self, data: Optional[list["Agent"]] = None):
        """Initialize a new AgentList.

        :param data: A list of Agents.
        """
        if data is not None:
            super().__init__(data)
        else:
            super().__init__()

    def shuffle(self, seed: Optional[str] = "edsl") -> AgentList:
        """Shuffle the AgentList.

        :param seed: The seed for the random number generator.
        """
        import random

        random.seed(seed)
        random.shuffle(self.data)
        return self

    def sample(self, n: int, seed: Optional[str] = None) -> AgentList:
        """Return a random sample of agents.

        :param n: The number of agents to sample.
        :param seed: The seed for the random number generator.
        """
        import random

        if seed:
            random.seed(seed)
        return AgentList(random.sample(self.data, n))

    def to_pandas(self):
        """Return a pandas DataFrame."""
        return self.to_scenario_list().to_pandas()

    def tally(self):
        return self.to_scenario_list().tally()

    def rename(self, old_name, new_name):
        """Rename a trait in the AgentList.

        :param old_name: The old name of the trait.
        :param new_name: The new name of the trait.
        """
        for agent in self.data:
            agent.rename(old_name, new_name)
        return self

    def select(self, *traits) -> AgentList:
        """Selects agents with only the references traits.

        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent(traits = {'a': 1, 'b': 1}), Agent(traits = {'a': 1, 'b': 2})])
        >>> al.select('a')
        AgentList([Agent(traits = {'a': 1}), Agent(traits = {'a': 1})])

        """

        if len(traits) == 1:
            traits_to_select = [list(traits)[0]]
        else:
            traits_to_select = list(traits)

        return AgentList([agent.select(*traits_to_select) for agent in self.data])

    def filter(self, expression: str) -> AgentList:
        """
        Filter a list of agents based on an expression.

        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent(traits = {'a': 1, 'b': 1}), Agent(traits = {'a': 1, 'b': 2})])
        >>> al.filter("b == 2")
        AgentList([Agent(traits = {'a': 1, 'b': 2})])
        """

        def create_evaluator(agent: "Agent"):
            """Create an evaluator for the given result.
            The 'combined_dict' is a mapping of all values for that Result object.
            """
            return EvalWithCompoundTypes(names=agent.traits)

        try:
            # iterates through all the results and evaluates the expression
            new_data = [
                agent for agent in self.data if create_evaluator(agent).eval(expression)
            ]
        except Exception as e:
            print(f"Exception:{e}")
            raise AgentListError(f"Error in filter. Exception:{e}")

        return AgentList(new_data)

    @property
    def all_traits(self):
        d = {}
        for agent in self:
            d.update(agent.traits)
        return list(d.keys())

    @classmethod
    def from_csv(cls, file_path: str, name_field: Optional[str] = None):
        """Load AgentList from a CSV file.

        >>> import csv
        >>> import os
        >>> with open('/tmp/agents.csv', 'w') as f:
        ...     writer = csv.writer(f)
        ...     _ = writer.writerow(['age', 'hair', 'height'])
        ...     _ = writer.writerow([22, 'brown', 5.5])
        >>> al = AgentList.from_csv('/tmp/agents.csv')
        >>> al
        AgentList([Agent(traits = {'age': '22', 'hair': 'brown', 'height': '5.5'})])
        >>> al = AgentList.from_csv('/tmp/agents.csv', name_field='hair')
        >>> al
        AgentList([Agent(name = \"""brown\""", traits = {'age': '22', 'height': '5.5'})])
        >>> os.remove('/tmp/agents.csv')

        :param file_path: The path to the CSV file.
        :param name_field: The name of the field to use as the agent name.
        """
        from edsl.agents.Agent import Agent

        agent_list = []
        with open(file_path, "r") as f:
            reader = csv.DictReader(f)
            for row in reader:
                if "name" in row:
                    import warnings

                    warnings.warn("Using 'name' field in the CSV for the Agent name")
                    name_field = "name"
                if name_field is not None:
                    agent_name = row.pop(name_field)
                    agent_list.append(Agent(traits=row, name=agent_name))
                else:
                    agent_list.append(Agent(traits=row))
        return cls(agent_list)

    def translate_traits(self, values_codebook: dict[str, str]):
        """Translate traits to a new codebook.

        :param codebook: The new codebook.
        """
        for agent in self.data:
            agent.translate_traits(codebook)
        return self

    def remove_trait(self, trait: str):
        """Remove traits from the AgentList.

        :param traits: The traits to remove.
        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent({'age': 22, 'hair': 'brown', 'height': 5.5}), Agent({'age': 22, 'hair': 'brown', 'height': 5.5})])
        >>> al.remove_trait('age')
        AgentList([Agent(traits = {'hair': 'brown', 'height': 5.5}), Agent(traits = {'hair': 'brown', 'height': 5.5})])
        """
        for agent in self.data:
            _ = agent.remove_trait(trait)
        return self

    def add_trait(self, trait, values):
        """Adds a new trait to every agent, with values taken from values.

        :param trait: The name of the trait.
        :param values: The valeues(s) of the trait. If a single value is passed, it is used for all agents.

        >>> al = AgentList.example()
        >>> al.add_trait('new_trait', 1)
        AgentList([Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5, 'new_trait': 1}), Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5, 'new_trait': 1})])
        >>> al.select('new_trait').to_scenario_list().to_list()
        [1, 1]
        >>> al.add_trait('new_trait', [1, 2, 3])
        Traceback (most recent call last):
        ...
        edsl.exceptions.agents.AgentListError: The passed values have to be the same length as the agent list.
        ...
        """
        if not is_iterable(values):
            value = values
            for agent in self.data:
                agent.add_trait(trait, value)
            return self

        if len(values) != len(self):
            raise AgentListError(
                "The passed values have to be the same length as the agent list."
            )
        for agent, value in zip(self.data, values):
            agent.add_trait(trait, value)
        return self

    @staticmethod
    def get_codebook(file_path: str):
        """Return the codebook for a CSV file.

        :param file_path: The path to the CSV file.
        """
        with open(file_path, "r") as f:
            reader = csv.DictReader(f)
            return {field: None for field in reader.fieldnames}

    def __hash__(self) -> int:
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self.to_dict(add_edsl_version=False, sorted=True))

    def to_dict(self, sorted=False, add_edsl_version=True):
        """Serialize the AgentList to a dictionary."""
        if sorted:
            data = self.data[:]
            data.sort(key=lambda x: hash(x))
        else:
            data = self.data

        d = {
            "agent_list": [
                agent.to_dict(add_edsl_version=add_edsl_version) for agent in data
            ]
        }
        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = "AgentList"

        return d

    def __eq__(self, other: AgentList) -> bool:
        return self.to_dict(sorted=True, add_edsl_version=False) == other.to_dict(
            sorted=True, add_edsl_version=False
        )

    def __repr__(self):
        return f"AgentList({self.data})"

    def _summary(self):
        return {
            "EDSL Class": "AgentList",
            "Number of agents": len(self),
            "Agent trait fields": self.all_traits,
        }

    def _repr_html_(self):
        """Return an HTML representation of the AgentList."""
        footer = f"<a href={self.__documentation__}>(docs)</a>"
        return str(self.summary(format="html")) + footer

    def to_csv(self, file_path: str):
        """Save the AgentList to a CSV file.

        :param file_path: The path to the CSV file.
        """
        self.to_scenario_list().to_csv(file_path)

    def to_list(self, include_agent_name=False) -> list[tuple]:
        """Return a list of tuples."""
        return self.to_scenario_list(include_agent_name).to_list()

    def to_scenario_list(self, include_agent_name=False) -> ScenarioList:
        """Return a list of scenarios."""
        from edsl.scenarios.ScenarioList import ScenarioList
        from edsl.scenarios.Scenario import Scenario

        if include_agent_name:
            return ScenarioList(
                [
                    Scenario(agent.traits | {"agent_name": agent.name})
                    for agent in self.data
                ]
            )
        return ScenarioList([Scenario(agent.traits) for agent in self.data])

    def table(
        self,
        *fields,
        tablefmt: Optional[str] = None,
        pretty_labels: Optional[dict] = None,
    ) -> Table:
        return (
            self.to_scenario_list()
            .to_dataset()
            .table(*fields, tablefmt=tablefmt, pretty_labels=pretty_labels)
        )

    def tree(self, node_order: Optional[List[str]] = None):
        return self.to_scenario_list().tree(node_order)

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data: dict) -> "AgentList":
        """Deserialize the dictionary back to an AgentList object.

        :param: data: A dictionary representing an AgentList.
        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent.example(), Agent.example()])
        >>> al2 = AgentList.from_dict(al.to_dict())
        >>> al2 == al
        True
        """
        from edsl.agents.Agent import Agent

        agents = [Agent.from_dict(agent_dict) for agent_dict in data["agent_list"]]
        return cls(agents)

    @classmethod
    def example(cls, randomize: bool = False) -> AgentList:
        """
        Returns an example AgentList instance.

        :param randomize: If True, uses Agent's randomize method.
        """
        from edsl.agents.Agent import Agent

        return cls([Agent.example(randomize), Agent.example(randomize)])

    @classmethod
    def from_list(self, trait_name: str, values: List[Any]):
        """Create an AgentList from a list of values.

        :param trait_name: The name of the trait.
        :param values: A list of values.

        >>> AgentList.from_list('age', [22, 23])
        AgentList([Agent(traits = {'age': 22}), Agent(traits = {'age': 23})])
        """
        from edsl.agents.Agent import Agent

        return AgentList([Agent({trait_name: value}) for value in values])

    def __mul__(self, other: AgentList) -> AgentList:
        """Takes the cross product of two AgentLists."""
        from itertools import product

        new_sl = []
        for s1, s2 in list(product(self, other)):
            new_sl.append(s1 + s2)
        return AgentList(new_sl)

    def code(self, string=True) -> Union[str, list[str]]:
        """Return code to construct an AgentList.

        >>> al = AgentList.example()
        >>> print(al.code())
        from edsl.agents.Agent import Agent
        from edsl.agents.AgentList import AgentList
        agent_list = AgentList([Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5}), Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5})])
        """
        lines = [
            "from edsl.agents.Agent import Agent",
            "from edsl.agents.AgentList import AgentList",
        ]
        lines.append(f"agent_list = AgentList({self.data})")
        if string:
            return "\n".join(lines)
        return lines

    def rich_print(self) -> Table:
        """Display an object as a rich table."""
        table = Table(title="AgentList")
        table.add_column("Agents", style="bold")
        for agent in self.data:
            table.add_row(agent.rich_print())
        return table


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import os
from collections import UserDict

from edsl.enums import service_to_api_keyname
from edsl.exceptions import MissingAPIKeyError


class KeyLookup(UserDict):
    @classmethod
    def from_os_environ(cls):
        """Create an instance of KeyLookupAPI with keys from os.environ"""
        return cls({key: value for key, value in os.environ.items()})

    def get_api_token(self, service: str, remote: bool = False):
        key_name = service_to_api_keyname.get(service, "NOT FOUND")

        if service == "bedrock":
            api_token = [self.get(key_name[0]), self.get(key_name[1])]
            missing_token = any(token is None for token in api_token)
        else:
            api_token = self.get(key_name)
            missing_token = api_token is None

        if missing_token and service != "test" and not remote:
            raise MissingAPIKeyError(
                f"""The key for service: `{service}` is not set.
                    Need a key with name {key_name} in your .env file."""
            )

        return api_token
